/*
 *	Copyright (c) 2000-2003 Rinet Corp., Novosibirsk, Russia
 *
 * Redistribution and use in source forms, with and without modification,
 * are permitted provided that this entire comment appears intact.
 *
 * THIS SOURCE CODE IS PROVIDED ``AS IS'' WITHOUT ANY WARRANTIES OF ANY KIND.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <syslog.h>
#include <ctype.h>
#include "netmon.h"

/*void inherit_var_data __P((VARIABLE *, VARIABLE *));*/
void splice_var_list __P((VARIABLE **, VARIABLE **));
int inherit_interface __P((INTERFACE *, INTERFACE *));
int splice_interface_list __P((INTERFACE **, INTERFACE **));
int inherit_bgpas __P((BGP_AS *, BGP_AS *));
int splice_bgpas_list __P((BGP_AS **, BGP_AS **));
int inherit_envmon __P((ENV_MON *, ENV_MON *));
int splice_envmon_list __P((ENV_MON **, ENV_MON **));
void inherit_object __P((OBJECT *, OBJECT *, OBJECT *));
OBJECT *splice_object_list __P((OBJECT *, OBJECT **, OBJECT **));
void free_interface __P((INTERFACE *));
void free_bgpas __P((BGP_AS *));
void free_envmon __P((ENV_MON *));

/* netmond.c */
extern CONFIG *cf;


static void
ptrswap(p1, p2)
	void **p1, **p2;
{
	void *p = *p1;
	*p1 = *p2;
	*p2 = p;
}

static void
fix_var_ref(obj)
	OBJECT *obj;
{
	METHOD *method;
	CHATSCRIPT *cs;
	REFERENCE *ref;

	for (method = obj->method_list; method; method = method->next) {
		for (cs = method->chatscript; cs; cs = cs->next) {
			for (ref = cs->var_ref; ref; ref = ref->next) {
				ref->var = get_var(obj->var_list, ref->name);
			}
		}
		if (method->snmpreq) {
			for (ref = method->snmpreq->var_ref; ref; ref = ref->next) {
				ref->var = get_var(obj->var_list, ref->name);
			}
		}
	}
}

#ifdef	unused
void
inherit_var_data(to, from)
	VARIABLE *to, *from;
{
	VARIABLE *curr, *var;

	for (curr = to; curr; curr = curr->next) {
		if ((var = get_var(from, curr->name)) != NULL) {
			curr->instance = var->instance;
			var->instance = NULL;
			curr->instance_len = var->instance_len;
			curr->mode = var->mode;
			curr->type = var->type;
			curr->value.integer = var->value.integer;
			var->value.integer = NULL;
			curr->length = var->length;
#ifdef	USE_VAR_MTIME
			curr->mtime = var->mtime;
#endif
			curr->statistic = var->statistic;
			var->statistic = NULL;
			curr->old = var->old;
			var->old = NULL;
		}
	}
}
#endif

void
splice_var_list(old, new)
	VARIABLE **old, **new;
{
	VARIABLE *curr = *old, *prev = NULL, *var, *temp;

	while (curr) {
		temp = NULL; /* as prev */
		for (var = *new; var; var = var->next) {
			if (!strcasecmp(curr->name, var->name))
				break;
			temp = var;
		}
		if (var) {
			if (var->value.integer &&
			    (!curr->value.integer ||
			     curr->length != var->length ||
			     memcmp(curr->value.integer, var->value.integer,
				    curr->length < var->length ?
				    curr->length : var->length))) {
				set_var(curr, var->type,
					var->value.integer, var->length);
			}
			if (var->objid &&
			    (!curr->objid ||
			     curr->objid_len != var->objid_len ||
			     memcmp(curr->objid, var->objid,
				    (curr->objid_len < var->objid_len ?
				     curr->objid_len : var->objid_len) *
				    sizeof(oid)))) {
				set_objid(curr, var->objid, var->objid_len);
			}
			if (temp)
				temp->next = var->next;
			else	*new = var->next;
			free_var(var);
		} else if (!curr->mode) {
			temp = curr->next;
			free_var(curr);
			if (prev)
				prev->next = temp;
			else	*old = temp;
			curr = temp;
			continue;
		}
		prev = curr;
		curr = curr->next;
	}
	if (*new) {	/* append rest of new to the end of old */
		if (prev)
			prev->next = *new;
		else	*old = *new;
		*new = NULL;	/* to avoid following frees */
	}
}

int
inherit_interface(old, new)
	INTERFACE *old, *new;
{
	int rescan = 0;

	ptrswap(&old->descr, &new->descr);
	ptrswap(&old->datadir, &new->datadir);
	ptrswap(&old->save_list, &new->save_list);
	ptrswap(&old->ns_acl, &new->ns_acl);

	rescan = (old->bandwidth != new->bandwidth); /* to force rescan */
	old->bandwidth = new->bandwidth;
	old->flags = new->flags;

	splice_var_list(&old->var_list, &new->var_list);

	return rescan;
}

int
splice_interface_list(old, new)
	INTERFACE **old, **new;
{
	INTERFACE *curr = *old, *prev = NULL, *iface, *temp;
	int rescan = 0;

	while (curr) {
		temp = NULL; /* as prev */
		for (iface = *new; iface; iface = iface->next) {
			if (curr->name) {
				if (iface->name &&
				    !strcasecmp(curr->name, iface->name))
					break;
			} else if (curr->index > 0) {
				if (iface->index > 0 &&
				    curr->index == iface->index)
					break;
			}
			temp = iface;
		}
		if (iface) {
			rescan += inherit_interface(curr, iface);
			if (temp)
				temp->next = iface->next;
			else	*new = iface->next;
			free_interface(iface);
			prev = curr;
			curr = curr->next;
		} else {
			rescan++; /* force rescan later */
			temp = curr->next;
			free_interface(curr);
			if (prev)
				prev->next = temp;
			else	*old = temp;
			curr = temp;
		}
	}

	if (*new) {	/* append rest of new to the end of old */
		rescan++;
		if (prev)
			prev->next = *new;
		else	*old = *new;
		*new = NULL;	/* to avoid following frees */
	}

	return rescan;
}

int
inherit_bgpas(old, new)
	BGP_AS *old, *new;
{
	ptrswap(&old->descr, &new->descr);
	ptrswap(&old->datadir, &new->datadir);
	ptrswap(&old->save_list, &new->save_list);
	ptrswap(&old->ns_acl, &new->ns_acl);

	if (!old->handconf && !new->handconf)
		return 0; /* still autodetect, no changes made */

	if (old->handconf && new->handconf) {
		/* still manualy configured -- check for bijection of peers */
		BGP_PEER *p1, *p2;

		for (p1 = old->peer; p1; p1 = p1->next) {
			for (p2 = new->peer; p2; p2 = p2->next) {
				if (!memcmp(&p1->ip_addr, &p2->ip_addr, sizeof(p1->ip_addr)))
					break;
			}
			if (!p2) break;
		}
		if (!p1) return 0; /* the same */
	}

	/* setup new list of peers */
	old->handconf = new->handconf;
	old->flags = new->flags;
	free_bgp_peers(old, 1);
	old->peer = new->peer;
	new->peer = NULL;

	return 1; /* rescan them to changes take effect */
}

int
splice_bgpas_list(old, new)
	BGP_AS **old, **new;
{
	BGP_AS *curr = *old, *prev = NULL, *bgp, *temp;
	int rescan = 0;

	while (curr) {
		temp = NULL; /* as prev */
		for (bgp = *new; bgp; bgp = bgp->next) {
			if (curr->asn == bgp->asn)
				break;
			temp = bgp;
		}
		if (bgp) {
			rescan += inherit_bgpas(curr, bgp);
			if (temp)
				temp->next = bgp->next;
			else	*new = bgp->next;
			free_bgpas(bgp);
			prev = curr;
			curr = curr->next;
		} else {
			temp = curr->next;
			free_bgpas(curr);
			if (prev)
				prev->next = temp;
			else	*old = temp;
			curr = temp;
		}
	}

	if (*new) {	/* append rest of new to the end of old */
		rescan++;
		if (prev)
			prev->next = *new;
		else	*old = *new;
		*new = NULL;	/* to avoid following frees */
	}

	return rescan;
}

int
inherit_envmon(old, new)
	ENV_MON *old, *new;
{
	ptrswap(&old->descr, &new->descr);
	ptrswap(&old->datadir, &new->datadir);
	ptrswap(&old->save_list, &new->save_list);
	ptrswap(&old->ns_acl, &new->ns_acl);

	if (!old->handconf && !new->handconf)
		return 0; /* still autodetect, no changes made */

	if (old->handconf && new->handconf) {
		/* still manualy configured -- check for bijection of gauges */

		if (old->type == new->type) {
			ENV_GAUGE *g1, *g2;
			for (g1 = old->gauge; g1; g1 = g1->next) {
				for (g2 = new->gauge; g2; g2 = g2->next) {
					if (!strcasecmp(g1->name, g2->name))
						break;
				}
				if (!g2) break;
			}
			if (!g1) return 0; /* the same */
		}
	}

	/* setup new list of peers */
	old->handconf = new->handconf;
	old->flags = new->flags;
	free_env_gauges(old, 1);
	old->gauge = new->gauge;
	new->gauge = NULL;

	return 1; /* rescan them to changes take effect */
}

int
splice_envmon_list(old, new)
	ENV_MON **old, **new;
{
	ENV_MON *curr = *old, *prev = NULL, *env, *temp;
	int rescan = 0;

	while (curr) {
		temp = NULL; /* as prev */
		for (env = *new; env; env = env->next) {
			if (curr->type == env->type)
				break;
			temp = env;
		}
		if (env) {
			rescan += inherit_envmon(curr, env);
			if (temp)
				temp->next = env->next;
			else	*new = env->next;
			free_envmon(env);
			prev = curr;
			curr = curr->next;
		} else {
			temp = curr->next;
			free_envmon(curr);
			if (prev)
				prev->next = temp;
			else	*old = temp;
			curr = temp;
		}
	}

	if (*new) {	/* append rest of new to the end of old */
		rescan++;
		if (prev)
			prev->next = *new;
		else	*old = *new;
		*new = NULL;	/* to avoid following frees */
	}

	return rescan;
}

void
inherit_object(parent, old, new)
	OBJECT *parent;
	OBJECT *old, *new;
{
	void *ip_addr, *ip_srcaddr;
	OBJECT *service;

	object_stop(old);

	ptrswap(&old->descr, &new->descr);
	ptrswap(&old->datadir, &new->datadir);
	ptrswap(&old->address, &new->address);
	if (parent) {
		ip_addr = &parent->ip_addr;
		ip_srcaddr = &parent->ip_srcaddr;
	} else {	
	    ip_addr = &new->ip_addr;
	    ip_srcaddr = &new->ip_srcaddr;
	}
	old->parent = parent;

	if (memcmp(&old->ip_addr, ip_addr, sizeof(old->ip_addr))) {
		memcpy(&old->ip_addr, ip_addr, sizeof(old->ip_addr));
		if (old->snmpdata) {
			free_var_list(old->snmpdata->ifTable);
			free_var_list(old->snmpdata->ipTable);
			free_var_list(old->snmpdata->bgpTable);
			free_var_list(old->snmpdata->envTable);
			memset(old->snmpdata, 0, sizeof(SNMP_DATA));
		}
	}
	if (memcmp(&old->ip_srcaddr, ip_srcaddr, sizeof(old->ip_srcaddr))) 
		memcpy(&old->ip_srcaddr, ip_srcaddr, sizeof(old->ip_srcaddr));

	old->polling = new->polling;
	old->saving = new->saving;
	old->sync = new->sync;
	old->timeout = new->timeout;
	old->retries = new->retries;

	ptrswap(&old->trap_list, &new->trap_list);
	ptrswap(&old->save_list, &new->save_list);
	ptrswap(&old->ns_acl, &new->ns_acl);
	ptrswap(&old->method_list, &new->method_list);

	splice_var_list(&old->var_list, &new->var_list);
	fix_var_ref(old);

	if (splice_interface_list(&old->interface, &new->interface) && old->snmpdata)
		old->snmpdata->pending_scan |= SCAN_IFTABLE;

	if (splice_interface_list(&old->ifgroup, &new->ifgroup) && old->snmpdata)
		old->snmpdata->pending_scan |= SCAN_IFTABLE;

	if (splice_bgpas_list(&old->bgp, &new->bgp) && old->snmpdata)
		old->snmpdata->pending_scan |= SCAN_BGPPEER;

	if (splice_envmon_list(&old->env, &new->env) && old->snmpdata)
		old->snmpdata->pending_scan |= SCAN_ENVMON;

	object_init(old);

	service = splice_object_list(old, &old->service, &new->service);
	for (; service; service = service->next) {
		service->ip_addr = old->ip_addr;
		service->ip_srcaddr = old->ip_srcaddr;
		service->flags = old->flags;
		service->parent = old;
		object_init(service);
	}
}

OBJECT *
splice_object_list(parent, old, new)
	OBJECT *parent;
	OBJECT **old, **new;
{
	OBJECT *curr = *old, *prev = NULL, *obj, *temp;
	int cnt = 0;

	while (curr) {
		temp = NULL; /* as prev */
		for (obj = *new; obj; obj = obj->next) {
			if (!strcasecmp(curr->name, obj->name))
				break;
			temp = obj;
		}
		if (obj) {	/* inherited object */
			if (temp)
				temp->next = obj->next;
			else
				*new = obj->next;
			if (parent == NULL && object_busy(curr)) {
				/* save object to the reconfig list */
				obj->next = cf->reconfig;
				cf->reconfig = obj;
				cnt++;
			} else {
				inherit_object(parent, curr, obj);
				obj->next = NULL; /* only one to free */
				free_object_list(obj, 1);
			}
			prev = curr;
			curr = curr->next;
		} else {	/* object was revoked from new config */
			temp = curr->next;
			curr->next = NULL; /* only one to free */
			free_object_list(curr, 0);
			if (prev)
				prev->next = temp;
			else
				*old = temp;
			curr = temp;
		}
	}

	if (*new) {	/* append rest of new to the end of old */
		curr = *new;
		if (prev)
			prev->next = *new;
		else	*old = *new;
		*new = NULL;	/* to avoid following frees */
	}
	if (cnt > 0)
		report(LOG_NOTICE, "soft_reconfig: %d objects were busy", cnt);
	return (curr);
}

int
rootdir_configure(CONFIG *cf)
{

	umask(cf->umask);
	mkdir(cf->rootdir, 0775);
	if (chdir(cf->rootdir) < 0) {
		report(LOG_ERR, "chdir \"%s\": %m", cf->rootdir);
		return (-1);
	}
	return (0);
}

void
finish_soft_reconfig(void)
{
	OBJECT *new, *obj, *prev, *tmp;

	if ((cf->flags & CONFIG_F_SOFTRECONFIG) == 0)
		return;

	prev = NULL;
	new = cf->reconfig;
	while (new) {
		for (obj = cf->target; obj; obj = obj->next) {
			if (strcasecmp(obj->name, new->name) == 0)
				break;
		}
		if (obj) {
			if (object_busy(obj)) {
				/* this object is still busy, try next */
				prev = new;
				new = new->next;
				continue;
			}
			inherit_object(NULL, obj, new);
			start_method_list(obj);
		}
		/* remove object from the reconfig list */
		tmp = new->next;
		new->next = NULL; /* only one to free */
		free_object_list(new, 1);
		if (prev != NULL)
			prev->next = tmp;
		new = tmp;
	}
	if (prev == NULL) {
		cf->reconfig = NULL;
		cf->flags &= ~CONFIG_F_SOFTRECONFIG;
		report(LOG_NOTICE, "soft reconfig completed");
	}
}

void
soft_reconfig(CONFIG *cf_new)
{
	dprintf(("soft_reconfig: entered\n"));

	umask(cf_new->umask);
	if (cf_new->rootdir && strcmp(cf->rootdir, cf_new->rootdir)) {
		mkdir(cf_new->rootdir, 0775);
		if (chdir(cf_new->rootdir) < 0) {
			report(LOG_ERR, "chdir \"%s\": %m", cf_new->rootdir);
			report(LOG_ERR, "soft reconfig failed");
			return;
		}
		ptrswap(&cf->rootdir, &cf_new->rootdir);
	}
	cf->umask = cf_new->umask;
	if (cf_new->rootdir) free(cf_new->rootdir);

	ptrswap(&cf->chrootdir, &cf_new->chrootdir);
	if (cf_new->chrootdir) free(cf_new->chrootdir);

	ptrswap(&cf->username, &cf_new->username);
	if (cf_new->username) free(cf_new->username);
	cf->uid = cf_new->uid;

	ptrswap(&cf->groupname, &cf_new->groupname);
	if (cf_new->groupname) free(cf_new->groupname);
	cf->gid = cf_new->gid;

	ptrswap(&cf->timefmt, &cf_new->timefmt);
	if (cf_new->timefmt) free(cf_new->timefmt);

	ptrswap(&cf->srcaddress, &cf_new->srcaddress);
	if (cf_new->srcaddress) free(cf_new->srcaddress);
	memcpy( &cf->ip_srcaddr, &cf_new->ip_srcaddr, sizeof(struct in_addr));
	cf->polling = cf_new->polling;
	cf->saving = cf_new->saving;
	cf->timeout = cf_new->timeout;
	cf->retries = cf_new->retries;
	cf->flags = cf_new->flags | CONFIG_F_SOFTRECONFIG;
	compatibility_flag = (cf->flags & CONFIG_F_COMPAT);

	if ((cf->enable_traps != cf_new->enable_traps) || memcmp(&cf->trap_ip_addr, &cf_new->trap_ip_addr, sizeof(struct in_addr)) ) {
		ptrswap(&cf->trap_address, &cf_new->trap_address);
		if (cf_new->trap_address) free(cf_new->trap_address);
		memcpy( &cf->trap_ip_addr, &cf_new->trap_ip_addr, sizeof(struct in_addr));
		cf->enable_traps = cf_new->enable_traps;
		trap_init(cf->enable_traps > 0);
	}
	cf->source_traps = cf_new->source_traps;

	if ((cf->ns_port != cf_new->ns_port) || memcmp(&cf->ns_ip_addr, &cf_new->ns_ip_addr, sizeof(struct in_addr))) {
		ptrswap(&cf->ns_address, &cf_new->ns_address);
		if (cf_new->ns_address) free(cf_new->ns_address);
		memcpy( &cf->ns_ip_addr, &cf_new->ns_ip_addr, sizeof(struct in_addr));
		cf->ns_port = cf_new->ns_port;
		netstate_init(cf->ns_port);
	}
	cf->ns_timo = cf_new->ns_timo;

	ptrswap(&cf->ns_acl, &cf_new->ns_acl);
	free_group_ref(&cf_new->ns_acl);

	ptrswap(&cf->group_list, &cf_new->group_list);
	free_group_list(&cf_new->group_list, 0);

	cf->reconfig = NULL;
	start_object_list(splice_object_list(NULL, &cf->target, &cf_new->target));
	if ((cf->flags & CONFIG_F_SOFTRECONFIG) == 0)
		report(LOG_NOTICE, "soft reconfig completed");
}

void
hard_reconfig(cf_new)
	CONFIG *cf_new;
{
	CONFIG *cf_cur, *cf_swap;

	dprintf(("hard_reconfig: entered\n"));

	if ((cf_cur = (CONFIG *)malloc(sizeof(CONFIG))) == NULL) {
		report(LOG_ERR, "malloc: Out of memory");
		return;
	}

	if (rootdir_configure(cf_new) < 0)
		return;

	memcpy(cf_cur, cf_new, sizeof(CONFIG));
	cf_swap = cf;
	cf = cf_cur;
	cf_cur = cf_swap;

#ifdef	DEBUG_MEMORY
	set_cksum();
#endif
	free_events();
	free_object_list(cf_cur->target, 1);

	if (cf_cur->rootdir) free(cf_cur->rootdir);
	if (cf_cur->chrootdir) free(cf_cur->chrootdir);
	if (cf_cur->username) free(cf_cur->username);
	if (cf_cur->groupname) free(cf_cur->groupname);
	if (cf_cur->srcaddress) free(cf_cur->srcaddress);
	if (cf_cur->ns_address) free(cf_cur->ns_address);
	if (cf_cur->trap_address) free(cf_cur->trap_address);
	if (cf_cur->timefmt) free(cf_cur->timefmt);

	free_group_ref(&cf_cur->ns_acl);
	free_group_list(&cf_cur->group_list, 0);
	free(cf_cur);

	trap_init(cf->enable_traps > 0);
	netstate_init(cf->ns_port);
	start_object_list(cf->target);
	report(LOG_NOTICE, "hard reconfig completed");
}

void
free_interface(iface)
	INTERFACE *iface;
{
	remove_event(NULL, iface);
	if (iface->name) free(iface->name);
	if (iface->descr) free(iface->descr);
	if (iface->datadir) free(iface->datadir);
	free_var_list(iface->var_list);
	free_save_list(iface->save_list);
	free_group_ref(&iface->ns_acl);
	free_ifaddr(&iface->ifaddr);
	free(iface);
}

void
free_bgpas(bgp)
	BGP_AS *bgp;
{
	remove_event(NULL, bgp);
	if (bgp->descr) free(bgp->descr);
	if (bgp->datadir) free(bgp->datadir);
	free_save_list(bgp->save_list);
	free_group_ref(&bgp->ns_acl);
	free_bgp_peers(bgp, 1);
	free(bgp);
}

void
free_envmon(env)
	ENV_MON *env;
{
	remove_event(NULL, env);
	if (env->descr) free(env->descr);
	if (env->datadir) free(env->datadir);
	free_save_list(env->save_list);
	free_group_ref(&env->ns_acl);
	free_env_gauges(env, 1);
	free(env);
}

void
free_object_list(OBJECT *object, int just_free)
{
	OBJECT *obj;
	INTERFACE *iface;
	BGP_AS *bgp;
	ENV_MON *env;
	void *next;

	/* Objects */
	obj = object;
	while (obj) {
		/* close all session methods at the object first */
		if (!just_free)
			remove_event(NULL, obj);
		free_method_list(obj->method_list);

		if (obj->name) free(obj->name);
		if (obj->descr) free(obj->descr);
		if (obj->datadir) free(obj->datadir);
		if (obj->address) free(obj->address);
		if (obj->srcaddress) free(obj->srcaddress);
		free_trap_list(obj->trap_list);
		free_var_list(obj->var_list);
		free_save_list(obj->save_list);
		free_group_ref(&obj->ns_acl);
		free_netpath(obj->nexthop);
		free_ifaddr(&obj->ifaddr);
		if (obj->snmpdata) {
			free_var_list(obj->snmpdata->ifTable);
			free_var_list(obj->snmpdata->ipTable);
			free_var_list(obj->snmpdata->bgpTable);
			free_var_list(obj->snmpdata->envTable);
			free(obj->snmpdata);
		}

		/* Interfaces */
		iface = obj->interface;
		while (iface) {
			next = iface->next;
			free_interface(iface);
			iface = next;
		}
		/* IfGroups */
		iface = obj->ifgroup;
		while (iface) {
			next = iface->next;
			free_interface(iface);
			iface = next;
		}

		/* BGP */
		bgp = obj->bgp;
		while (bgp) {
			next = bgp->next;
			free_bgpas(bgp);
			bgp = next;
		}

		/* Environment monitors */
		env = obj->env;
		while (env) {
			next = env->next;
			free_envmon(env);
			env = next;
		}

		/* Subobjects */
		free_object_list(obj->service, just_free);

		next = obj->next;
		free(obj);
		obj = next;
	}
}
